# -*- coding: utf-8 -*-

CARDINAL_POINTS = ["N", "E", "S", "W"]

class Orientation:
    """
    This class represent the orientation
    Args:
        cardinal_letter: str
            corresponding to the upper letter of the cardinal point
            => accepted values : N, E, S, W
    """

    def __init__(self, cardinal_letter: str):
        if cardinal_letter.upper() not in CARDINAL_POINTS:
            print("====", cardinal_letter)
            raise ValueError("Not a cardinal point, accepted values => N, E, S, W")
        self._current = cardinal_letter

    def rotate_right(self):
        """
        TODO
        """
        self._current = CARDINAL_POINTS[(CARDINAL_POINTS.index(self._current) + 1) % 4]

    def rotate_left(self):
        """
        TODO
        """
        self._current = CARDINAL_POINTS[(CARDINAL_POINTS.index(self._current) - 1) % 4]

    def is_north(self) -> bool:
        """
        return True if the current orientation is the north
        """
        return self._current == "N"

    def is_east(self) -> bool:
        """
        return True if the current orientation is the east
        """
        return self._current == "E"

    def is_south(self) -> bool:
        """
        return True if the current orientation is the south
        """
        return self._current == "S"

    def is_west(self) -> bool:
        """
        return True if the current orientation is the west
        """
        return self._current == "W"

    def __str__(self):
        return self._current



# -*- coding: utf-8 -*-

from copy import deepcopy
from time import sleep

from .coordinates import Coordinates
from .lawn import Lawn
from .orientation import Orientation


class Mower:
    """
    This class contain the coordinates and the orientation of a mower.
    A mower is able to turn left, turn right and go forward
    Args:
        coordinates: Coordinates
            coordinates of the mower on the lawn
        orientation: str
            letter corresponding to one of the 4 cardinal points (N, E, S, W)
    Attributes:
        coordinates: Coordinates
            coordinates of the mower on the lawn
        orientation: Orientation
            orientation of the mower, a letter corresponding to one of the 4 cardinal points (N, E, S, W)
    """

    def __init__(
        self,
        lawn: Lawn,
        coordinates: Coordinates,
        orientation: Orientation,
        instructions: "list[str]",
    ):
        self.uid: str = f"{coordinates}{orientation}".replace(" ", "")
        self.lawn: Lawn = lawn
        self.coordinates: Coordinates = coordinates
        self.orientation: Orientation = orientation
        self.instructions: list[str] = instructions

    def start_mowing(self):
        """
        Execute the mowing instructions
        """
        for instruction in self.instructions:
            if instruction == "L":
                self.turn_left()
            elif instruction == "R":
                self.turn_right()
            elif instruction == "F":
                self.move_forward()
            sleep(1)  # TODO move to config

    def turn_left(self):
        """
        Rotate the mower of 90° at the left
        """
        previous_orientation = deepcopy(self.orientation)
        self.orientation.rotate_left()
        print(
            f"Mower {self.uid} has turned to left (from {previous_orientation} to {self.orientation})"
        )

    def turn_right(self):
        """
        Rotate the mower of 90° at the right
        """
        previous_orientation = deepcopy(self.orientation)
        self.orientation.rotate_right()
        print(
            f"Mower {self.uid} has turned to right (from {previous_orientation} to {self.orientation})"
        )

    def move_forward(self):
        """
        Move forward the mower
        """
        new_coordinates = deepcopy(self.coordinates)
        if self.orientation.is_north():
            new_coordinates.y += 1
        elif self.orientation.is_east():
            new_coordinates.x += 1
        elif self.orientation.is_south():
            new_coordinates.y -= 1
        elif self.orientation.is_west():
            new_coordinates.x -= 1
        if self.lawn.is_valid_coordinates(new_coordinates):
            if self.lawn.is_available_coordinates(new_coordinates):
                self.coordinates = deepcopy(new_coordinates)
                print(
                    f"Mower {self.uid} has moved forward in the direction of the cardinal point {self.orientation} (new coordinates :{self.coordinates})"
                )
            else:
                print(
                    f"Mower {self.uid} the new coordinates are bussy, the instructions are ingored"
                )
        else:
            print(
                f"Mower {self.uid} the new coordinates are outside of the lawn, the instructions are ingored"
            )

    def __str__(self) -> str:
        """
        Return a string corresponding to position of the mower,
        """
        return f"{self.coordinates} {self.orientation}"



# -*- coding: utf-8 -*-

import sys
from threading import Thread

from .core import Config, Lawn, Mower


def wait_all_mowers(mowing_list: "list[Thread]"):
    """
    Wait for all mowers to finish mowing
    """
    for thread in mowing_list:
        thread.join()
    mowing_list.clear()


def start(file_path):
    """
    TODO
    """
    config = Config(file_path)
    lawn = Lawn(config.upper_right_corner)
    for mower in config.mower_list:
        lawn.mowers.append(Mower(lawn, mower[0], mower[1], mower[2]))

    mowing_list: list[Thread] = []
    for mower in lawn.mowers:
        mowing = Thread(target=mower.start_mowing)
        mowing_list.append(mowing)
        mowing.start()
    wait_all_mowers(mowing_list)
    for mower in lawn.mowers:
        print(mower)


if __name__ == "__main__":
    if len(sys.argv) == 1:
        print("You must provide an input file !")
    else:
        start(sys.argv[1])



# -*- coding: utf-8 -*-

from typing import Tuple

from .coordinates import Coordinates
from .orientation import Orientation


class Config:
    def __init__(self, file_path: str) -> None:
        # TODO check if file existe
        self.lignes: list[str] = self.__parse_file__(file_path)
        self.upper_right_corner: Coordinates = self.__get_upper_right_corner__()
        self.mower_list: list[Tuple(Coordinates, Orientation, Tuple[str])] = []
        self.__parse_mower_list__()
        pass

    def __parse_file__(self, file_path: str):
        with open(file_path, "r") as f:
            return f.read().splitlines()

    def __get_upper_right_corner__(self) -> Tuple[int, int]:
        first_line_data = self.lignes.pop(0).split(" ")
        # TODO check data
        return Coordinates(int(first_line_data[0]), int(first_line_data[1]))

    def __parse_mower_list__(self) -> None:
        for a, b in zip(*[iter(self.lignes)] * 2):
            coordinates = self.__parse_mower_coordinates__(a)
            orientation = self.__parse_mower_orientation__(a)
            instructions = self.__parse_mower_instructions__(b)
            self.mower_list.append((coordinates, orientation, instructions))
        return

    def __parse_mower_coordinates__(self, position: str) -> Coordinates:
        return Coordinates(int(position.split(" ")[0]), int(position.split(" ")[1]))

    def __parse_mower_orientation__(self, position: str) -> Orientation:
        return Orientation(position.split(" ")[2])

    def __parse_mower_instructions__(self, instructions: str) -> Tuple[str]:
        return tuple(instructions)



# -*- coding: utf-8 -*-

class Coordinates:
    """
    This class represent coordinates points
    """

    def __init__(self, x: int, y: int):
        self.x: int = x
        self.y: int = y

    def __eq__(self, other): 
        if not isinstance(other, Coordinates):
            return NotImplemented

        return self.x == other.x and self.y == other.y

    def __le__(self, other):
        if not isinstance(other, Coordinates):
            return NotImplemented

        return self.x <= other.x and self.y <= other.y

    def __str__(self):
        return f"{self.x} {self.y}"


# -*- coding: utf-8 -*-

from .coordinates import Coordinates

BOTTOM_LEFT_CORNER = Coordinates(0, 0)


class Lawn:
    """
    Lawn class
    ==========
    This class represent a lawn surface

    To simplify the navigation, the surface is a grid

    Args:
         upper_right_corner: Tuple[int, int]
            coordinates (x, y) of the cell from the upper right corner of the grid

    Attributes:
        dimensions: Tule[Coordinates, Coordinates]
            dimensions of the grid
        mowers: list[Mower]
            collection of mowers

    Methodes:
        TODO
    """

    def __init__(self, upper_right_corner: Coordinates):
        self.dimensions = (BOTTOM_LEFT_CORNER, upper_right_corner)
        self.mowers: list = list()

    def is_available_coordinates(self, coordinates: Coordinates) -> bool:
        """
        TODO
        """
        return coordinates not in [m.coordinates for m in self.mowers]

    def is_valid_coordinates(self, coordinates: Coordinates) -> bool:
        """
        TODO
        """
        return BOTTOM_LEFT_CORNER <= coordinates <= self.dimensions[1]
